/* ****************************************************************
   File: VLMMHeaderC.h
   Author: Kingsley Sage, University of Sussex, Brighton, UK
   Description: Contains performance controlling parameters
   Version: Release version 1.0

   This code is supplied with no warranty whatsoever and was developed
   with support from the EU ActIPret project  and in fulfilment of 
   Kingsley Sage's DPhil work on 'Task Based Visual Control'. Any comments
   or suggestions should be e-mailed to khs20@sussex.ac.uk.

   The code may be reproduced freely for non-commercial use provided
   that re-use acknowledges Kingsley Sage.
   ***************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <stdlib.h>
#include <float.h>

#define NODE_E -1
#define NODE_NULL -2
#define MAX_L 13              /* Max depth of prediction suffix tree */
#define MAX_LENGTH_SIGMA 10   /* Max numbers of symbols in the alphabet */
#define TRUE 1
#define FALSE 0
#define PI 3.14159265359

/* ==========================================================
   Two global variables: */

int LengthSigma;
int ncells; /* Number of cells in Data for learn and number of cells in Tree for conv2MM */
int ncellsT;
int F; /* Number of feature dimensions */

/* ========================================================== */

struct QQ
  {
    double *base; /* Use calloc to create enough room for the matrix */
    int mrows;
    int ncols;
  };

struct trans_s
  {
     int nodeLength;
     int node[MAX_L];
  };

struct TreeElement
  {
     int node[MAX_L];
     int nodeLength;
     int parent[MAX_L];
     int parentLength;
     double PTable[MAX_LENGTH_SIGMA];
     int terminal; /* Actually a bool, only used in conv2MM_C */
     struct trans_s trans[MAX_LENGTH_SIGMA]; /* Only used in conv2MM_C */
     struct TreeElement *next;
  };

struct SElement
  {
     int node[MAX_L];
     int nodeLength;
     struct SElement *next;
  };

struct ModelElement
  {
    int node[MAX_L];
    int nodeLength;
    int ref;
    struct trans_s trans[MAX_LENGTH_SIGMA];
    double PTable[2][MAX_LENGTH_SIGMA];
    struct ModelElement *next;
  };

struct RefinedModelElement
  {
    int node[MAX_L];
    int nodeLength;
    int ref;
    struct trans_s trans[MAX_LENGTH_SIGMA];
    double PTable[2][MAX_LENGTH_SIGMA];
    double *mu;  /* Ptr to an array of mu values */
    double *cov; /* Ptr to a 2D covariance matrix */
    double *covInv; /* Ptr to inverse of 2D covariance matrix */
    double covInvDet; /* Determinant of inverse of 2D covariance matrix */
    struct RefinedModelElement *next;
  };


/* Macros to look up entries in Data by cell,row and column  ... */
#define elM(Data,cell,row,col) *((Data+(cell-1))->base+((Data+(cell-1))->mrows*(col-1)+(row-1)))

#define el(Data,cell,row,col) *((Data+(cell))->base+((Data+(cell))->mrows*(col)+(row)))

/* END OF FILE */
