/* ****************************************************************
   File: conv2MM_C.c
   Author: Kingsley Sage, University of Sussex, Brighton, UK
   Description: Converts a prediction suffix tree to a probabilistic
                finite automata
   Version: Release version 1.0

   This code is supplied with no warranty whatsoever and was developed
   with support from the EU ActIPret project  and in fulfilment of 
   Kingsley Sage's DPhil work on 'Task Based Visual Control'. Any comments
   or suggestions should be e-mailed to khs20@sussex.ac.uk.

   The code may be reproduced freely for non-commercial use provided
   that re-use acknowledges Kingsley Sage.
   ***************************************************************/

#include <mex.h>
#include "VLMMHeaderC.h"
#include "conv2MM_Cimpl.c"

/* The gateway function */
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
int 		mrows,ncols;
int 		i,j,i1,z,ModelLength;
double          *tmpdblPtr;
struct TreeElement *Tree, *TreePtr;
struct ModelElement *Model,*ModelPtr;

int             nfields,ModelFields;
const char      **fnames, **Modelfnames;
mxArray         *entry,*cellArrayPtr;
mxArray         *tmp1,*tmp3,*tmp5,*tmp7,*tmp9;
double          *tmp2,*tmp4,*tmp6,*tmp10;
int cellArrayCol;


nfields   = 3;
fnames    = mxCalloc(nfields,sizeof(char *));
fnames[0] = mxMalloc(sizeof(char)*10);
fnames[1] = mxMalloc(sizeof(char)*10);
fnames[2] = mxMalloc(sizeof(char)*10);
strcpy(fnames[0],"node\0");
strcpy(fnames[1],"parent\0");
strcpy(fnames[2],"PTable\0");
ModelFields    = 4;
Modelfnames    = mxCalloc(ModelFields,sizeof(char *));
Modelfnames[0] = mxMalloc(sizeof(char)*10);
Modelfnames[1] = mxMalloc(sizeof(char)*10);
Modelfnames[2] = mxMalloc(sizeof(char)*10);
Modelfnames[3] = mxMalloc(sizeof(char)*10);
strcpy(Modelfnames[0],"node\0");
strcpy(Modelfnames[1],"ref\0");
strcpy(Modelfnames[2],"trans\0");
strcpy(Modelfnames[3],"PTable\0");

if (nrhs!=2)
  mexErrMsgTxt("Need Tree and LengthSigma !!");
if (!mxIsDouble(prhs[1])||mxIsComplex(prhs[1]))
  mexErrMsgTxt("LengthSigma must be an int");
LengthSigma=floor(*mxGetPr(prhs[1]));
if (LengthSigma<=0)
  mexErrMsgTxt("LengthSigma must be > 0");
if (LengthSigma>MAX_LENGTH_SIGMA)
  mexErrMsgTxt("Need to recompile with bigger value for MAX_LENGTH_SIGMA");

ncells  = mxGetNumberOfElements(prhs[0]);
Tree    = (struct TreeElement *) mxCalloc(ncells,sizeof(struct TreeElement));
TreePtr = Tree;

/* Now decode Tree from MATLAB space into a linked list ... */
/* The list also happens here to be in contiguous memory */

for (i=0;i<ncells;i++)
  {
    tmp3  = mxGetCell(prhs[0],i);
    tmp1  = mxGetFieldByNumber(tmp3,0,mxGetFieldNumber(tmp3,fnames[0]));
    ncols = mxGetN(tmp1);
    (TreePtr+i)->nodeLength = ncols;
    tmpdblPtr = mxCalloc(ncols,sizeof(double));
    tmpdblPtr = mxGetPr(tmp1);
    for (j=0;j<ncols;j++)
      (TreePtr+i)->node[j] = (int) floor(*(tmpdblPtr+j));

    tmp1  = mxGetFieldByNumber(tmp3,0,mxGetFieldNumber(tmp3,fnames[1]));
    ncols = mxGetN(tmp1);
    (TreePtr+i)->parentLength = ncols;
    tmpdblPtr = mxCalloc(ncols,sizeof(double));
    tmpdblPtr = mxGetPr(tmp1);
    for (j=0;j<ncols;j++)
      (TreePtr+i)->parent[j] = (int) floor(*(tmpdblPtr+j));
    
    tmp1      = mxGetFieldByNumber(tmp3,0,mxGetFieldNumber(tmp3,fnames[2]));
    mrows     = mxGetM(tmp1);
    tmpdblPtr = mxCalloc(mrows,sizeof(double));
    tmpdblPtr = mxGetPr(tmp1);
    for (j=0;j<mrows;j++)
      (TreePtr+i)->PTable[j] = *(tmpdblPtr+j);
    
    if (i<(ncells-1))
      (TreePtr+i)->next = TreePtr+i+1;
    else
      (TreePtr+i)->next = NULL;
  }


/* Now do the hard work ... */

Model = conv2MM_Cimpl(Tree);

/* Now decode the Model from the C function back to MATLAB ... */

ModelLength=0;
for (ModelPtr=Model;ModelPtr!=NULL;ModelPtr=ModelPtr->next)
  ModelLength++;
mexPrintf("Model Length = %d\n",ModelLength);

cellArrayPtr=mxCreateCellMatrix(1,ModelLength);

cellArrayCol=0;

for (ModelPtr=Model;ModelPtr!=NULL;ModelPtr=ModelPtr->next)
  {
    entry = mxCreateStructMatrix(1,1,ModelFields,Modelfnames);
    tmp1  = mxCreateDoubleMatrix(1,ModelPtr->nodeLength,mxREAL);
    tmp2  = mxCalloc(ModelPtr->nodeLength,sizeof(double));
    for (z=0;z<ModelPtr->nodeLength;z++)
      *(tmp2+z)=(double)(ModelPtr->node[z]);
    memcpy(mxGetPr(tmp1),tmp2,ModelPtr->nodeLength*sizeof(double));  
    mxSetFieldByNumber(entry,0,mxGetFieldNumber(entry,Modelfnames[0]),tmp1);

    tmp3  = mxCreateDoubleMatrix(1,1,mxREAL);
    tmp4  = mxMalloc(sizeof(double));
    *tmp4 = (double)(ModelPtr->ref)+1.0;
    memcpy(mxGetPr(tmp3),tmp4,sizeof(double));
    mxSetFieldByNumber(entry,0,mxGetFieldNumber(entry,Modelfnames[1]),tmp3);

    tmp5  = mxCreateCellMatrix(1,LengthSigma);
    for (i1=0;i1<LengthSigma;i1++)
      {
        tmp7 = mxCreateDoubleMatrix(1,ModelPtr->trans[i1].nodeLength,mxREAL);
        tmp6 = mxCalloc(ModelPtr->trans[i1].nodeLength,sizeof(double));
        for (z=0;z<ModelPtr->trans[i1].nodeLength;z++)
          *(tmp6+z)=(double)(ModelPtr->trans[i1].node[z]);
        memcpy(mxGetPr(tmp7),tmp6,ModelPtr->trans[i1].nodeLength*sizeof(double));
        mxSetFieldByNumber(tmp5,0,i1,tmp7);
      }
    mxSetFieldByNumber(entry,0,mxGetFieldNumber(entry,Modelfnames[2]),tmp5);

    tmp9  = mxCreateDoubleMatrix(LengthSigma,2,mxREAL);
    tmp10 = mxCalloc(LengthSigma*2,sizeof(double));
    for (z=0;z<LengthSigma;z++)
      {
        *(tmp10+z)=(double)(ModelPtr->PTable[0][z]);
        *(tmp10+z+LengthSigma)=(double)(ModelPtr->PTable[1][z])+1.0;
      }
    memcpy(mxGetPr(tmp9),tmp10,LengthSigma*2*sizeof(double));
    mxSetFieldByNumber(entry,0,mxGetFieldNumber(entry,Modelfnames[3]),tmp9);

    mxSetCell(cellArrayPtr,cellArrayCol,entry);
    cellArrayCol++;

  }

plhs[0]=cellArrayPtr;

}

/* END OF FILE */
