/* ****************************************************************
   File: conv2MM_Cimpl.c
   Author: Kingsley Sage, University of Sussex, Brighton, UK
   Description: Main body of code for PST to PFA conversion
   Version: Release version 1.0

   This code is supplied with no warranty whatsoever and was developed
   with support from the EU ActIPret project and in part fulfilment of 
   Kingsley Sage's DPhil work on 'Task Based Visual Control'. Any comments
   or suggestions should be e-mailed to khs20@sussex.ac.uk.

   The code may be reproduced freely for non-commercial use provided
   that re-use acknowledges Kingsley Sage.
   ***************************************************************/

int NodeCmp(int *, int *, int);
int NodeCmp2(int *, int *, int, int);
void CopyNodeValues(struct TreeElement *, struct TreeElement *);

int NodeCmp(int *List1, int *List2, int Length)
/* Compares digits in List1 to List2 up to Length. Returns TRUE iff the same, FALSE otherwise */
{
int i,stop;

stop=FALSE;
for (i=0;((i<Length)&&(stop==FALSE));i++)
  {
    if (List1[i]!=List2[i]) stop=TRUE;
  }
if (stop==TRUE) return FALSE;
else return TRUE;;
}


int NodeCmp2(int *List1, int *List2, int Length1, int Length2)
/* Compares Length1 digits in List1 to last Length1 digits in List2. Returns TRUE iff the same, FALSE otherwise */
{
int i, stop;

stop=FALSE;
for (i=0;((i<Length1)&&(stop==FALSE));i++)
  {
    if (List1[i]!=List2[Length2-Length1+i]) stop = TRUE;
  }
if (stop==TRUE) return FALSE;
else return TRUE;
}

void CopyNodeValues(struct TreeElement *t1, struct TreeElement *t2)
/* Copies values from t1 to t2 */
{
  int i;
  t2->nodeLength=t1->nodeLength;
  for (i=0;i<t1->nodeLength;i++)
    t2->node[i]=t1->node[i];
  t2->parentLength=t1->parentLength;
  for (i=0;i<t1->parentLength;i++)
    t2->parent[i]=t1->parent[i];
  for (i=0;i<LengthSigma;i++)
    t2->PTable[i]=t1->PTable[i];
  t2->terminal=t1->terminal;

}

struct ModelElement * conv2MM_Cimpl(struct TreeElement *Tree)
{
struct TreeElement *TreePtr,*TreePtr2,*NewTreeNodePtr, *LeafSet,*LeafSetPtr,*LeafSetPtr2,*NewLeafNodePtr;
struct TreeElement *SortedLeafSet, *SortedLeafSetPtr;
struct ModelElement *Model, *ModelPtr;
int i1,i2,i3,i4,i5,i6,i7,i8;
int KeepLooking, FoundShorterPrefix;
int LengthT,Depth,LengthSortedLeafSet;
int SLSIndex;
int LenthSortedLeafSet, LengthPostFix, FoundTransition, LengthComparedTo;
int PostFix[MAX_L+1],Prefix[MAX_L+1], ShorterPrefix[MAX_L];
int NodesAdded,stop,FoundNode,FoundNodeFrom,FoundShorterNode,NoInwardTransition;
int PrefixLength, ShorterPrefixLength, Offset;
struct trans_s LookingFor;
FILE *op;
int SaveIt;

NodesAdded=0;
SaveIt=FALSE;

/* Establish the length of T */
LengthT=0;
for (TreePtr=Tree;TreePtr!=NULL;TreePtr=TreePtr->next)
  LengthT++;
mexPrintf("LengthT = %d\n",LengthT);

mexPrintf("Building a list of leaf nodes ...\n");
Tree->terminal = FALSE;
for (i1=1;i1<LengthT;i1++)
  {
    KeepLooking = TRUE;
    i2=1;
    while ((KeepLooking==TRUE)&&(i2<LengthT))
      {
        if ((Tree+i1)->nodeLength==(Tree+i2)->parentLength)
          {
            if (NodeCmp((Tree+i1)->node,(Tree+i2)->parent,(Tree+i1)->nodeLength))
              {
                KeepLooking = FALSE;
                (Tree+i1)->terminal = FALSE;
              }
          }
        i2++;
      }
    if (KeepLooking==TRUE)
      (Tree+i1)->terminal=TRUE;
  }

/* Now we need to know the actual depth of T */
Depth = 0;
for (i1=0;i1<LengthT;i1++)
  {
    if ((Tree+i1)->nodeLength>Depth)
      Depth=(Tree+i1)->nodeLength;
  }
mexPrintf("Depth = %d\n",Depth);

redo:
/* Here we have to treat Tree as a proper linked list ... */

mexPrintf("Redo\n");
/* And finally build the list of leaf nodes ... */
LeafSet = NULL;
for (TreePtr=Tree->next;TreePtr!=NULL;TreePtr=TreePtr->next)
/* Excludes 'e' ... */
  {
    if (TreePtr->terminal==TRUE)
      {
        NewLeafNodePtr = (struct TreeElement *) mxMalloc(sizeof(struct TreeElement));
        CopyNodeValues(TreePtr,NewLeafNodePtr);
        NewLeafNodePtr->next = NULL;
        if (LeafSet==NULL)
          LeafSet=NewLeafNodePtr;
        else
          {
            for (LeafSetPtr=LeafSet;LeafSetPtr->next!=NULL;LeafSetPtr=LeafSetPtr->next);
            LeafSetPtr->next=NewLeafNodePtr;
          }
      }
  }

/* Now order the LeafSet in descending size of node label length */
/* As we know the size of SortedLeafSet, set it up as a contiguous memory block */
i1=0;
for (LeafSetPtr=LeafSet;LeafSetPtr!=NULL;LeafSetPtr=LeafSetPtr->next)
  i1++;
SortedLeafSet=(struct TreeElement *) mxCalloc(i1,sizeof(struct TreeElement));
LengthSortedLeafSet = i1;
/* Now add the link pointers (just for the sake of completeness ...) */
for (i2=0;i2<(i1-1);i2++)
  (SortedLeafSet+i2)->next=(SortedLeafSet+i2+1);
(SortedLeafSet+i1-1)->next=NULL;

mexPrintf("Sorting leaf set into descending order ...\n");
SLSIndex=0;
for (i1=Depth;i1>0;i1--)
  {
    for (LeafSetPtr=LeafSet;LeafSetPtr!=NULL;LeafSetPtr=LeafSetPtr->next)
      {
        if (LeafSetPtr->nodeLength==i1)
          {
            CopyNodeValues(LeafSetPtr,(SortedLeafSet+SLSIndex));
            for (i3=0;i3<LengthSigma;i3++)
              {
                (SortedLeafSet+SLSIndex)->trans[i3].nodeLength=0;
                for (i4=0;i4<MAX_L;i4++)
                  (SortedLeafSet+SLSIndex)->trans[i3].node[i4]=0;
              }
            SLSIndex++;
          }
      }
  }
mexPrintf("Done ...\n");

mexPrintf("Checking for extra nodes to account for entrance transitions ...\n");


i1=0;
while ((i1<LengthSortedLeafSet)&&((SortedLeafSet+i1)->nodeLength>1))
  {
    /* First check to see if there is an indirect transition into the
       node indexed by i1 */

    i2=0;
    FoundNodeFrom=FALSE;
    while ((i2<LengthSortedLeafSet)&&(FoundNodeFrom==FALSE))
      {
        i3=1;
        while ((i3<=LengthSigma)&&(FoundNodeFrom==FALSE))
          {
            for (i4=0;i4<(SortedLeafSet+i2)->nodeLength;i4++) 
              PostFix[i4]=(SortedLeafSet+i2)->node[i4];
            PostFix[(SortedLeafSet+i2)->nodeLength]=i3;
            LengthPostFix=(SortedLeafSet+i2)->nodeLength + 1;
            LengthComparedTo=(SortedLeafSet+i1)->nodeLength;
            if (LengthPostFix==LengthComparedTo)
              {
                if (NodeCmp((SortedLeafSet+i1)->node,PostFix,LengthPostFix))
                  {
                     FoundNodeFrom=TRUE;
                  }
              }
            else if (LengthPostFix>LengthComparedTo)
              {
                if (NodeCmp2((SortedLeafSet+i1)->node,PostFix,LengthComparedTo,LengthPostFix))
                  {
                    FoundNodeFrom=TRUE;
                  }
              }
            i3++;
          }
        i2++;
      } 

    /* If no such transition exists, look to see if the prefix of the
       node indexed by i1 is contained anywhere in the SortedLeafSet */
    if (FoundNodeFrom==FALSE)
      {
        PrefixLength=(SortedLeafSet+i1)->nodeLength-1;
        for (i3=0;i3<PrefixLength;i3++) Prefix[i3]=(SortedLeafSet+i1)->node[i3];
        i2=0;
        while ((FoundNodeFrom==FALSE)&&(i2<LengthSortedLeafSet))
          {
            if ((SortedLeafSet+i2)->nodeLength==PrefixLength)
              {
                if (NodeCmp(Prefix,(SortedLeafSet+i2)->node,PrefixLength))
                  FoundNodeFrom=TRUE;
              }
            i2++;
          }
       }

    /* If FoundNodeFrom is now still FALSE, there is no possible transition 
       into (SortedLeafSet+i1)->node. In this case, we need to
       expand a lower level node to provide a prefix for (SortedLeafSet+i1)->node */
    PrefixLength=(SortedLeafSet+i1)->nodeLength-1;
    for (i7=0;i7<PrefixLength;i7++) Prefix[i7]=(SortedLeafSet+i1)->node[i7];
    if (FoundNodeFrom==FALSE)
      {
        mexPrintf("No inwards transition for:\n");
        for (i7=0;i7<(SortedLeafSet+i1)->nodeLength;i7++)
          mexPrintf("%d ",(SortedLeafSet+i1)->node[i7]);
        mexPrintf("\n\n");

        /* Look for the longest node in SortedLeafSet that can act as 
           a prefix. We know that Prefix itself is not present
           in the SortedLeafSet and that there has to be a prefix in there
           at some depth (even if it's length 1) ... */
        ShorterPrefixLength=PrefixLength-1;
        FoundShorterNode=FALSE;
        NoInwardTransition=FALSE;
        while ((FoundShorterNode==FALSE)&&(NoInwardTransition==FALSE))
          {
            Offset=PrefixLength-ShorterPrefixLength;
            for (i4=0;i4<ShorterPrefixLength;i4++) ShorterPrefix[i4]=Prefix[i4+Offset];
            i3=0;
            while ((FoundShorterNode==FALSE)&&(i3<LengthSortedLeafSet))
              {
                if ((SortedLeafSet+i3)->nodeLength==ShorterPrefixLength)
                  {
                    if (NodeCmp((SortedLeafSet+i3)->node,ShorterPrefix,ShorterPrefixLength))
                      FoundShorterNode=TRUE;
                  }
                if (FoundShorterNode==FALSE) i3++;
              }
            if (FoundShorterNode==FALSE) ShorterPrefixLength--;
            if (ShorterPrefixLength==0)
              {
                mexPrintf("Still no inwards transition for:\n");
                for (i7=0;i7<(SortedLeafSet+i1)->nodeLength;i7++)
                  mexPrintf("%d ",(SortedLeafSet+i1)->node[i7]);
                mexPrintf("\n\n");
                FoundNodeFrom=FALSE; /* The node will never be accessed ... */
                FoundShorterNode=FALSE;
                NoInwardTransition=TRUE;
              }
          }

        if ((FoundShorterNode==TRUE)&&(NoInwardTransition==FALSE))
          {
            mexPrintf("Prefix To be expanded:\n");
            for (i7=0;i7<ShorterPrefixLength;i7++)
              mexPrintf("%d ",ShorterPrefix[i7]);
            mexPrintf("\n");

            /* Now ShorterPrefix needs expanding to Prefix ... */
            for (TreePtr=Tree;TreePtr->next!=NULL;TreePtr=TreePtr->next);
            /* TreePtr now points to the last item in the list */
            i8=ShorterPrefixLength+1;
            for (i6=i8;i6<=PrefixLength;i6++)
              {
mexPrintf("i6 = %d\n",i6);
                /* i6 is the depth of the new nodes that we need to add ... */
                /* First mark the ShorterPrefix as a non-terminal node. We can't just
                   use the previous i3 index because it will change in subsequent iterations
                   to the correct prefix depth ...*/
                FoundShorterPrefix=FALSE;
                TreePtr2=Tree;
                while (FoundShorterPrefix==FALSE)
                /* Replaced the old for statement here which seemed to not work properly */
                  {
                    if (TreePtr2->nodeLength==ShorterPrefixLength)
                      {
                        if (NodeCmp(TreePtr2->node,ShorterPrefix,ShorterPrefixLength))
                          {
                            FoundShorterPrefix=TRUE;
                            TreePtr2->terminal=FALSE;
                          }
                      }
                    if (FoundShorterPrefix==FALSE)
                      TreePtr2=TreePtr2->next;
                  }

                for (i4=1;i4<=LengthSigma;i4++)
                  {
                    NewTreeNodePtr=(struct TreeElement *) mxMalloc(sizeof(struct TreeElement));
                    for (i5=0;i5<ShorterPrefixLength;i5++)
                      NewTreeNodePtr->node[i5+1]=ShorterPrefix[i5];
                    NewTreeNodePtr->node[0]=i4;
                    NewTreeNodePtr->nodeLength=ShorterPrefixLength+1;
                    for (i5=0;i5<ShorterPrefixLength;i5++)
                      NewTreeNodePtr->parent[i5]=ShorterPrefix[i5];
                    NewTreeNodePtr->parentLength=ShorterPrefixLength;
                    NewTreeNodePtr->terminal=TRUE;

                    for (i5=0;i5<LengthSigma;i5++)
                      NewTreeNodePtr->PTable[i5]=TreePtr2->PTable[i5];
                    /* All these nodes will have the same PTable entries */
                    NewTreeNodePtr->next=NULL;
                    TreePtr->next=NewTreeNodePtr;
                    TreePtr=NewTreeNodePtr;
                    NodesAdded++;
                    LengthT++;
                    mexPrintf("Added node:\n");
                    for (i7=0;i7<NewTreeNodePtr->nodeLength;i7++)
                      mexPrintf("%d ",NewTreeNodePtr->node[i7]);
                    mexPrintf("\n\n");
                  }

                /* Now extend ShorterPrefix by the next symbol in Prefix so that the
                   i6 loop continues to build the prefix suffix tree in the right way ... */
                ShorterPrefixLength++;
                Offset=PrefixLength-ShorterPrefixLength;
                for (i4=0;i4<ShorterPrefixLength;i4++)
                  ShorterPrefix[i4]=Prefix[i4+Offset];              
              }
          }      
        /* Then escape from the i1 loop .... */
        if (FoundShorterNode==TRUE)
          i1=LengthSortedLeafSet+1;

      }
    i1++;
  }

if ((FoundNodeFrom==FALSE)&&(NoInwardTransition==FALSE))
  {
    /* Free up the old SortedLeafSet (contiguous block) and the LeafSet (linked list) */
    mxFree(SortedLeafSet);
    LeafSetPtr=LeafSet;
    while(LeafSetPtr!=NULL)
      {
        LeafSetPtr2=LeafSetPtr;
        LeafSetPtr=LeafSetPtr->next;
        mxFree(LeafSetPtr2);
      }
    SortedLeafSet=NULL;
    LeafSet=NULL;
  }

if ((FoundNodeFrom==FALSE)&&(NoInwardTransition==FALSE)) goto redo;
if (FoundNodeFrom==FALSE) mexPrintf("Buggered it up ...\n");

mexPrintf("Setting up node transition lists ... \n");

i1=0;
while (i1<LengthSortedLeafSet)
  {
    i2=1;
    while (i2<=LengthSigma)
      {
        i3=0;
        for (i4=0;i4<(SortedLeafSet+i1)->nodeLength;i4++) 
          PostFix[i4]=(SortedLeafSet+i1)->node[i4];
        PostFix[(SortedLeafSet+i1)->nodeLength]=i2;
        LengthPostFix=(SortedLeafSet+i1)->nodeLength + 1;
        FoundTransition=FALSE;
        while ((FoundTransition==FALSE)&&(i3<LengthSortedLeafSet))
          {
            LengthComparedTo=(SortedLeafSet+i3)->nodeLength;
            if (LengthPostFix==LengthComparedTo)
              {
                if (NodeCmp((SortedLeafSet+i3)->node,PostFix,LengthPostFix))
                  {
                    FoundTransition=TRUE;
                    (SortedLeafSet+i1)->trans[i2-1].nodeLength=(SortedLeafSet+i3)->nodeLength;
                    for (i4=0;i4<(SortedLeafSet+i3)->nodeLength;i4++)
                      (SortedLeafSet+i1)->trans[i2-1].node[i4]=(SortedLeafSet+i3)->node[i4];
                  }
              }
            else if (LengthPostFix>LengthComparedTo)
              {
                if (NodeCmp2((SortedLeafSet+i3)->node,PostFix,LengthComparedTo,LengthPostFix))
                  {
                    FoundTransition=TRUE;
                    (SortedLeafSet+i1)->trans[i2-1].nodeLength=(SortedLeafSet+i3)->nodeLength;
                    for (i4=0;i4<(SortedLeafSet+i3)->nodeLength;i4++)
                      (SortedLeafSet+i1)->trans[i2-1].node[i4]=(SortedLeafSet+i3)->node[i4];
                  }
              }
            i3++;
          }

        if (FoundTransition==FALSE)
          {
            mexPrintf("No exit transition for ...\n");
            for (i4=0;i4<(SortedLeafSet+i1)->nodeLength;i4++)
            mexPrintf("%d ",(SortedLeafSet+i1)->node[i4]);
            mexPrintf("with symbol %d\n",i2);
            /* So set the trans entry to -2 so we can deal with this later ... */
            (SortedLeafSet+i1)->trans[i2-1].nodeLength=1;
            (SortedLeafSet+i1)->trans[i2-1].node[0]=-1; 
          }
        i2++;
      }
    i1++;
  }

mexPrintf("Nodes added = %d\n",NodesAdded);

mexPrintf("Converting to bounded MM ...\n");

/* Now build the final bounded MM */
Model=(struct ModelElement *) mxCalloc(LengthSortedLeafSet,sizeof(struct ModelElement));
/* Now add the link pointers (just for the sake of completeness ...) */
i1=LengthSortedLeafSet;
for (i2=0;i2<(i1-1);i2++)
  (Model+i2)->next=(Model+i2+1);
(Model+i1-1)->next=NULL;

i1=0;
for (i1=0;i1<LengthSortedLeafSet;i1++)
  {
    i3=(SortedLeafSet+i1)->nodeLength;
    (Model+i1)->nodeLength=(SortedLeafSet+i1)->nodeLength;
    for (i4=0;i4<i3;i4++)
      (Model+i1)->node[i4]=(SortedLeafSet+i1)->node[i4];
    for (i4=0;i4<LengthSigma;i4++)
      {
        (Model+i1)->trans[i4].nodeLength=(SortedLeafSet+i1)->trans[i4].nodeLength;
        for (i5=0;i5<(SortedLeafSet+i1)->trans[i4].nodeLength;i5++)
          (Model+i1)->trans[i4].node[i5]=(SortedLeafSet+i1)->trans[i4].node[i5];
      }
    for (i4=0;i4<LengthSigma;i4++)
      {
        (Model+i1)->PTable[0][i4] = (SortedLeafSet+i1)->PTable[i4];
        (Model+i1)->PTable[1][i4] = 0;
      }
    (Model+i1)->ref=i1;
  }

/* Done with the Leaf Sets now, so delete them ... */
/* Needs updating ... */
mxFree(SortedLeafSet);
for (LeafSetPtr=LeafSet;LeafSetPtr!=NULL;LeafSetPtr=LeafSetPtr->next)
  mxFree(LeafSetPtr);
SortedLeafSet=NULL;
LeafSet=NULL;

for (i1=0;i1<LengthSortedLeafSet;i1++)
  {
    for (i2=0;i2<LengthSigma;i2++)
      {
        if ((Model+i1)->trans[i2].node[0]!=-1)
          {
            i3=0;
            LookingFor.nodeLength=(Model+i1)->trans[i2].nodeLength;
            for (i4=0;i4<LookingFor.nodeLength;i4++)
              LookingFor.node[i4]=(Model+i1)->trans[i2].node[i4];
            FoundNode=FALSE;
            while ((i3<LengthSortedLeafSet)&&(FoundNode==FALSE))
              {
                if ((Model+i3)->nodeLength==LookingFor.nodeLength)
                  {
                    if (NodeCmp((Model+i3)->node,LookingFor.node,LookingFor.nodeLength))
                      {
                        FoundNode=TRUE;
                        (Model+i1)->PTable[1][i2]=(Model+i3)->ref;
                      }
                  }
                i3++;
              }
          }
        else
          {
            (Model+i1)->PTable[1][i2]=-2;
            /* (Model+i1)->PTable[0][i2]=0.0; */
          }
      }
  }

end:

return Model;

}

/* END OF FILE */
