/* ****************************************************************
   File: learnC.c
   Author: Kingsley Sage, University of Sussex, Brighton, UK
   Description: mex implementation of the original learn.m function
   Version: Release version 1.0

   This code is supplied with no warranty whatsoever and was developed
   with support from the EU ActIPret project and in part fulfilment of 
   Kingsley Sage's DPhil work on 'Task Based Visual Control'. Any comments
   or suggestions should be e-mailed to khs20@sussex.ac.uk.

   The code may be reproduced freely for non-commercial use provided
   that re-use acknowledges Kingsley Sage.
   ***************************************************************/

#include <mex.h>
#include "VLMMHeaderC.h"
#include "learnCimpl.c"

/* The gateway function */
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
int 		mrows,ncols;
int 		i;
int 		n_incell;
struct QQ 	*Data;
double 		MinE;
int 		L;
struct TreeElement *Tree,*TreePtr;

int             nfields;
const char      **fnames;
mxArray         *entry, *tmp,*tmp3,*tmp5, *cellArrayPtr;
int             cellArrayCol;
double          *tmp2,*tmp4,*tmp6;
int             z,TreeLength;

if (nrhs!=3)
  mexErrMsgTxt("Need Data, MinE and L !!");

ncells = mxGetNumberOfElements(prhs[0]);
Data   = (struct QQ *) mxCalloc(ncells,sizeof(struct QQ));
mrows  = mxGetM(prhs[1]);
ncols  = mxGetN(prhs[1]);
if (!mxIsDouble(prhs[1])||mxIsComplex(prhs[1])||!(mrows==1&&ncols==1))
  mexErrMsgTxt("MinE must be a non-complex scalar double");
MinE   = *mxGetPr(prhs[1]);
mrows  = mxGetM(prhs[2]);
ncols  = mxGetN(prhs[2]);
if (!mxIsDouble(prhs[2])||mxIsComplex(prhs[2])||!(mrows==1&&ncols==1))
  mexErrMsgTxt("L must be an int");
L=floor(*mxGetPr(prhs[2]));
if (L<0)
  mexErrMsgTxt("L must be > 0");

for (i=0;i<ncells;i++)
  {
    (Data+i)->mrows = mxGetM(mxGetCell(prhs[0],i));
    (Data+i)->ncols = mxGetN(mxGetCell(prhs[0],i));
    (Data+i)->base  = mxGetPr(mxGetCell(prhs[0],i));
  }

/* Now do the hard work ... */
LengthSigma=mxGetM(mxGetCell(prhs[0],0));
if (LengthSigma>MAX_LENGTH_SIGMA)
  mexErrMsgTxt("Need to recompile with bigger value for MAX_LENGTH_SIGMA");
if (L>MAX_L)
  mexErrMsgTxt("Need to recompile with bigger value for MAX_L");

Tree = learnCimpl(Data,MinE,L);

/* Now decode the Tree from the C function back to MATLAB ... */
TreeLength=0;
for (TreePtr=Tree;TreePtr!=NULL;TreePtr=TreePtr->next)
  TreeLength++;
nfields=3;
fnames    = mxCalloc(nfields,sizeof(char *));
fnames[0] = mxMalloc(sizeof(char)*10);
fnames[1] = mxMalloc(sizeof(char)*10);
fnames[2] = mxMalloc(sizeof(char)*10);
strcpy(fnames[0],"node\0");
strcpy(fnames[1],"parent\0");
strcpy(fnames[2],"PTable\0");

cellArrayPtr=mxCreateCellMatrix(1,TreeLength);

cellArrayCol=0;

for (TreePtr=Tree;TreePtr!=NULL;TreePtr=TreePtr->next)
  {
    entry = mxCreateStructMatrix(1,1,nfields,fnames);
    tmp   = mxCreateDoubleMatrix(1,TreePtr->nodeLength,mxREAL);
    tmp2  = mxCalloc(TreePtr->nodeLength,sizeof(double));
    for (z=0;z<TreePtr->nodeLength;z++)
      *(tmp2+z)=(double)(TreePtr->node[z]) +1.0;
    memcpy(mxGetPr(tmp),tmp2,TreePtr->nodeLength*sizeof(double));  
    mxSetFieldByNumber(entry,0,mxGetFieldNumber(entry,fnames[0]),tmp);

    if (TreePtr->parentLength>0)
      tmp3=mxCreateDoubleMatrix(1,TreePtr->parentLength,mxREAL);
    else tmp3=mxCreateDoubleMatrix(0,0,mxREAL);
    tmp4=mxCalloc(TreePtr->parentLength,sizeof(double));
    for (z=0;z<TreePtr->parentLength;z++)
      *(tmp4+z)=(double)(TreePtr->parent[z]) +1.0;
    memcpy(mxGetPr(tmp3),tmp4,TreePtr->parentLength*sizeof(double));
    mxSetFieldByNumber(entry,0,mxGetFieldNumber(entry,fnames[1]),tmp3);
      
    tmp5  = mxCreateDoubleMatrix(LengthSigma,1,mxREAL);
    tmp6  = mxCalloc(LengthSigma,sizeof(double));
    for (z=0;z<LengthSigma;z++)
      *(tmp6+z)=(double)(TreePtr->PTable[z]);
    memcpy(mxGetPr(tmp5),tmp6,LengthSigma*sizeof(double));
    mxSetFieldByNumber(entry,0,mxGetFieldNumber(entry,fnames[2]),tmp5);

    mxSetCell(cellArrayPtr,cellArrayCol,entry);
    cellArrayCol++;
  }

plhs[0]=cellArrayPtr;

/* And clean up ... */

}

/* END OF FILE */
