/* ****************************************************************
   File: learnCimpl.c
   Author: Kingsley Sage, University of Sussex, Brighton, UK
   Description: Body of implementation of mex function
   Version: Release version 1.1

   This code is supplied with no warranty whatsoever and was developed
   with support from the EU ActIPret project  and in fulfilment of 
   Kingsley Sage's DPhil work on 'Task Based Visual Control'. Any comments
   or suggestions should be e-mailed to khs20@sussex.ac.uk.

   The code may be reproduced freely for non-commercial use provided
   that re-use acknowledges Kingsley Sage.
   ***************************************************************/

double p1(struct QQ *,int *,int);
double p2(struct QQ *,int, int *,int);
double err(struct QQ *, int *,int);
int NodeCmp(int *, int *, int);
struct SElement * AddNodeToS(struct SElement *, int *, int);
struct TreeElement * learnC(struct QQ *,double,int);

double p1(struct QQ *Data,int *nodeListP,int nodeLength)
/* Calculate p(s) */
{
double p,pqq;
int qq,i,i1;
double Denominator,Numerator,SubProduct;
int LengthDataqq;

p=0.0;
for (qq=0;qq<ncells;qq++)
  {
    LengthDataqq=(Data+qq)->ncols;
    Denominator=LengthDataqq - nodeLength + 1;
    Numerator=0.0;
    for (i=0;i<(LengthDataqq-nodeLength+1);i++)
      {
        SubProduct=1.0;
        for (i1=0;i1<nodeLength;i1++)
          {
            SubProduct=SubProduct*el(Data,qq,*(nodeListP+i1),i+i1);
          }
        Numerator=Numerator+SubProduct;
      }
    pqq = Numerator/Denominator;
    p = p+(pqq/ncells);
  }
return p;
}

double p2(struct QQ *Data, int symbol, int *sP, int sLength)
/* Calculates p(sigma|s) */
{
double p,pqq;
int qq,i,i1,j,z;
double Denominator,Numerator,SubProduct;
int LengthDataqq;
double total[MAX_LENGTH_SIGMA];
int BigString[MAX_L];
int BigStringLength;

p=0.0;
for (qq=0;qq<ncells;qq++)
  {
    LengthDataqq=(Data+qq)->ncols;
    for(z=0;z<LengthSigma;z++) total[z]=0.0;
    for (j=0;j<LengthSigma;j++)
      {
        /* Copy sP into front digits of BigString ... */
        for (z=0;z<sLength;z++) BigString[z]=*(sP+z);
        /* Add symbol onto the end of BigString ... */
        BigString[sLength]=j;
        BigStringLength=sLength+1;
        for (i=0;i<(LengthDataqq-BigStringLength+1);i++)
          {
            SubProduct=1.0;
            for (i1=0;i1<BigStringLength;i1++)
              {
                SubProduct=SubProduct*el(Data,qq,BigString[i1],i+i1);
              }
            total[j]=total[j]+SubProduct;
          }
      }
    Numerator=total[symbol];
    Denominator=0.0;
    for (z=0;z<LengthSigma;z++) Denominator=Denominator+total[z];
    if ((Denominator==0.0)||(Numerator==0.0)) pqq=0.0;
    else pqq=Numerator/Denominator;
    p = p+(pqq/ncells);
  }

return p;
}

double err(struct QQ *Data, int *sP, int sLength)
/* Calculates KL err for sp[0] vs. sp[1] */
{
double z,x1,x2,x3;
int i,j,kkLength;
int kk[MAX_L];

z=0.0;
x3=p1(Data,sP,sLength); /* Does not depend on i */
for (i=0;i<LengthSigma;i++)
  {
    for (j=0;j<sLength;j++) kk[j]=sP[j];
    kk[sLength]=i;
    kkLength=sLength+1;
    x1=p1(Data,kk,kkLength);
    if (x1>0.0)
      {
        x2=p2(Data,i,sP+1,sLength-1);
        z=z+(x1*log10(x1/(x2*x3)));
      }
  }
return z;
}


int NodeCmp(int *List1, int *List2, int Length)
/* Compares digits in List1 to List2 up to Length. Returns 1 iff the same, 0 otherwise */
{
int i,stop;

stop=0;
for (i=0;((i<Length)&&(stop==0));i++)
  {
    if (List1[i]!=List2[i]) stop=1;
  }
if (stop==1) return 0;
else return 1;
}


struct SElement *AddNodeToS(struct SElement *S,int *nodeListP,int nodeLength)
{
int i;
struct SElement *newPtr,*endPtr;

if (S==NULL)
/* This is the first entry in S */
  {
    S = mxMalloc(sizeof(struct SElement));
    for (i=0;i<nodeLength;i++)
      S->node[i]=*(nodeListP+i);
    S->nodeLength=nodeLength;
    S->next=NULL;
  }
else
  {
    newPtr=mxMalloc(sizeof(struct SElement));
    for (i=0;i<nodeLength;i++)
      newPtr->node[i]=*(nodeListP+i);
    newPtr->nodeLength=nodeLength;
    newPtr->next=NULL;
    for (endPtr=S;(endPtr->next!=NULL);endPtr=endPtr->next);
    endPtr->next=newPtr;
  }

return S;
}


/* The main file containing the C version of the learn routine */
struct TreeElement * learnCimpl(struct QQ *Data,double MinE, int L)
{
struct TreeElement *Tree,*TreePtr,*NewT;
int TreeLength;
struct SElement *S,*SPtr,*SOldPtr;
struct SElement PickS;
int i,i1,i3,i4,i5,i6;
int nodeList[MAX_L];
int add_node, add_sub_node;
int *kk,kkk;

for (i=0;i<MAX_L;i++) nodeList[i]=NODE_NULL;
S=NULL;

/* Create the root node 'e' */
Tree = mxMalloc(sizeof(struct TreeElement));
TreeLength = 1;
Tree->node[0]=NODE_E;
Tree->nodeLength=1;
Tree->parent[0]=NODE_NULL;
Tree->parentLength=0;
Tree->next=NULL;

for (i=0;i<LengthSigma;i++)
  {
    nodeList[0]=i;
    Tree->PTable[i]=p1(Data,nodeList,Tree->nodeLength);
    if (Tree->PTable[i]>=MinE)
      {
        /* Add the length 1 node to the candidate set S */
        S = AddNodeToS(S,nodeList,1);
      }
  }


/* Couple of p2 tests */
kkk=2;kk=&kkk;
mexPrintf("p2(2,3)=%f",p2(Data, 1, kk, 2));

/* Now the main part of the algorithm */
while (S!=NULL)
  {
    /* Take the last member of S, copy it to PickS and remove from S ... */
    SOldPtr=S;
    for (SPtr=S;(SPtr->next!=NULL);)
      {
        SOldPtr=SPtr;
        SPtr=SPtr->next;
      }
    PickS.nodeLength = SPtr->nodeLength;
    for (i=0;i<PickS.nodeLength;i++) PickS.node[i] = SPtr->node[i];
    PickS.next = NULL;
    SOldPtr->next = NULL;
    if (SPtr!=S) 
      { 
        mxFree(SPtr);
        SPtr = NULL;
      }
     else 
      {
        mxFree(S);
        S = NULL;
      }

    add_node=0;
    if (PickS.nodeLength==1)
      add_node=1;
    else if (err(Data,PickS.node,PickS.nodeLength)>=MinE)
      add_node=1;
    else
      add_node=0;

    if (add_node==1) /* Add node(s) corresponding to PickS to T */
      {
        for (i1=0;i1<PickS.nodeLength;i1++)
          {
            add_sub_node=1;
            for (TreePtr=Tree->next;TreePtr!=NULL;TreePtr=TreePtr->next)
              {
                if (TreePtr->nodeLength==(i1+1))
                  {
                    if (NodeCmp(TreePtr->node,&(PickS.node[PickS.nodeLength-i1-1]),TreePtr->nodeLength))
                      {
                        add_sub_node=0;
                      }
                  }
               }
            if (add_sub_node==1)
              {
                for (i4=0;i4<LengthSigma;i4++)
                  {
                    /* First check to see whether [i4,PickS] appears in Data */
                    nodeList[0]=i4;
                    i6=1;
                    for (i5=PickS.nodeLength-i1;i5<PickS.nodeLength;i5++)
                      nodeList[i6++]=PickS.node[i5];
                    if (p1(Data,nodeList,i6)>0.0)
                      {
                        TreeLength++;
                        NewT=mxMalloc(sizeof(struct TreeElement));
                        NewT->next=NULL;
                        for(TreePtr=Tree;TreePtr->next!=NULL;TreePtr=TreePtr->next);
                        TreePtr->next=NewT;
                        NewT->node[0]=i4;
                        i6=1;
                        for (i5=PickS.nodeLength-i1;i5<PickS.nodeLength;i5++) 
                          NewT->node[i6++]=PickS.node[i5];
                        NewT->nodeLength=i6;
                        if (PickS.nodeLength==1)
                          {
                            NewT->parent[0]=NODE_E;
                            NewT->parentLength=1;
                          }
                        else
                          {
                            i6=0;
                            for (i5=PickS.nodeLength-i1;i5<PickS.nodeLength;i5++)
                              NewT->parent[i6++]=PickS.node[i5];
                            NewT->parentLength=i6;
                          }
                        for (i3=0;i3<LengthSigma;i3++)
                          NewT->PTable[i3]=p2(Data,i3,NewT->node,NewT->nodeLength);
                      }
                  }
              }
          }
      }
    
    /* If |PickS| < L, then ... */
    if (PickS.nodeLength < L)
      {
        for (i4=0;i4<LengthSigma;i4++)
          {
            nodeList[0]=i4;
            i6=1;
            for (i5=0;i5<PickS.nodeLength;i5++)
              nodeList[i6++]=PickS.node[i5];
            if (p1(Data,nodeList,PickS.nodeLength+1)>=MinE)
              S=AddNodeToS(S,nodeList,PickS.nodeLength+1);
          }
      }
  }

end:

mexPrintf("TreeSize=%d\n",TreeLength);
return Tree; 
}

/* END OF FILE */
